--  title: SPACE
--     by: Matthew Sephton, Tektonix Software 
-- e-mail: u5ms@csc.liv.ac.uk
--    www: http://www.csc.liv.ac.uk/~u5ms
--    ed.: 0.20; FPS speed throttling, more transparent text display
--    ed.: 0.21; poke of screen bugs removed
--
-- Here's my first dabble in Euphoria, just to see what sort of speed it can
-- get up.  Use the numeric pad to control your flight through space (NUM LOCK
-- should be off) and feel free to change the constant parameters at the
-- beginning.  The speed is only really affected by those marked with a *
-- 
-- FUTURE: Totally free movement, planets, a spaceship, page swapping, a game?

without type_check

include graphics.e
include get.e
include machine.e

constant GRAPHICS_MODE = 19             -- 320x200 256 colours, see graphics.e
constant VGA_MEM = #A0000               -- start of VGA memory

constant POPULATION = 300               -- number of stars *
constant LEVELS = 6                     -- parallax levels (mults of 3 best)
constant MESSAGE = "Euphoria!"          -- message to display *
constant FPS = 1 / 50                   -- frames per sec (change divisor) *
constant text_y = 13                    -- vertical centre of screen

sequence stars, speeds, colours, dir
integer max_x, max_y, screen_x, screen_y, text_x
sequence vc

procedure init()
-- initialize global variables
    if graphics_mode(GRAPHICS_MODE) then
       puts(1, "Needs at least VGA graphics!\n")
       abort(1)
    end if
    vc = video_config()

    -- screen edge limits
    max_x = vc[VC_XPIXELS]
    max_y = vc[VC_YPIXELS]
    screen_x = max_x - 1
    screen_y = max_y - 1

    -- blue message in middle of screen
    text_color(BRIGHT_BLUE)
    text_x = floor((vc[VC_COLUMNS] / 2) - floor((length(MESSAGE)*2-1) / 2)) + 1

    stars = {}
    speeds = {}
    colours = {}
    dir = {1, 0}    -- stars going right initially
    
    -- speed throttle
    tick_rate(100)
end procedure

procedure phat(sequence loc, sequence s)
-- put a sequence at location {x,y} padded with gaps
    for c = 1 to length(s) do
        position(loc[2], loc[1]+2*(c-1))
        puts(1, s[c])
    end for
end procedure

procedure space()
-- main routine
    sequence old_pos
    integer x, y, s
    atom k, throttle

    init()

    -- main loop
    while 1 do
        k = get_key()
        if k != -1 then         -- key pressed...
           if k = 27 then       -- ESC, quits
              exit
           elsif k = 327 then   -- HOME, stars go down & left
              dir = {1, 1}
           elsif k = 328 then   -- UP, stars go down
              dir = {0, 1}
           elsif k = 329 then   -- PG UP, stars go down & right
              dir = {-1, 1}
           elsif k = 331 then   -- LEFT, stars go right
              dir = {1, 0}
           elsif k = 332 then   -- 5, stars stay still
              k = wait_key()
              -- dir = {0, 0}
           elsif k = 333 then   -- RIGHT, stars go left
              dir = {-1, 0}
           elsif k = 335 then   -- END, stars go up & left
              dir = {1, -1}
           elsif k = 336 then   -- DOWN, stars go up
              dir = {0, -1}
           elsif k = 337 then   -- PG DN, stars go up & right
              dir = {-1, -1}
           end if
        end if

        -- create starfield
        if length(stars) <= POPULATION then
            -- add a new star
            x = rand(screen_x)                          -- random position
            y = rand(screen_y)
            stars = append(stars, {x, y})               
            s = rand(LEVELS)                            -- random level
            speeds = append(speeds, s)

            if s > floor(LEVELS * 2/3) then             -- top (white)
               colours = append(colours, BRIGHT_WHITE)
            elsif s <= floor(LEVELS / 3) then           -- bottom (grey)
               colours = append(colours, GRAY)
            else                                        -- middle (light grey)
               colours = append(colours, WHITE)
            end if
        end if
        
        -- speed throttle
        throttle = time()
        -- move all the stars
        for i = 1 to length(stars) do
            old_pos = stars[i]
            stars[i][1] = stars[i][1] + (dir[1] * speeds[i]) 
            stars[i][2] = stars[i][2] + (dir[2] * speeds[i]) 

            -- check for off screen location
            if (dir[1] = -1 and stars[i][1] < 0) then
               stars[i][1] = max_x + stars[i][1]
            elsif (dir[1] = 1 and stars[i][1] > screen_x) then
               stars[i][1] = vc[VC_XPIXELS] - stars[i][1]
            elsif (dir[2] = -1 and stars[i][2] < 0) then
               stars[i][2] = max_y + stars[i][2]
            elsif (dir[2] = 1 and stars[i][2] > screen_y) then
               stars[i][2] = vc[VC_YPIXELS] - stars[i][2]
            end if
            
            -- *twice* as fast as pixel() on my P166
            -- blank out previous position
            poke(VGA_MEM + old_pos[1]+old_pos[2]*max_x, BLACK)
            -- draw at new position
            poke(VGA_MEM + stars[i][1]+stars[i][2]*max_x, colours[i])
        end for
        
        -- message
        phat({text_y, text_x}, MESSAGE)
        
        -- speed throttle
        while time()-throttle < FPS do
        end while
    end while
    
    --normal tick rate
    tick_rate(0)    
    -- revert to old graphics mode
    if graphics_mode(-1) then
    end if
end procedure

-- start program
space()

